#ifndef _mpm_
#define _mpm_

#include "grid.h"
#include "file_io.h"
#include <limits>
#include "Boundaries.h"

using namespace GRIDS;
using namespace ALGEBRA;

template<class T>
class MPM_DRIVER{
	
	int m,n;//m and n are the number of cells per dimension in the grid
	T dx,dt,dt_per_frame,dt_target,time,dx_particle,CFL;
	GRID_2D<T>& node_grid;
	VECTOR<T>& particle_position;
	VECTOR<T>& particle_velocity;
	VECTOR<T>& particle_mass;
	VECTOR<T>& particle_area;
	
	VECTOR< MATRIX_2X2<T> > F;		//Deformation gradient.
	VECTOR< MATRIX_2X2<T> > FE;		//Elastic deformation gradient.
	VECTOR< MATRIX_2X2<T> > FP;		//Plastic deformation gradient.

	T thetaC;						//Compression limit constant.
	T thetaS;						//Stretching limit constant.
	T hardening;					//Hardening dimensionless constant.

	VECTOR<T> eulerian_velocity,eulerian_velocity_old;
	VECTOR<T> eulerian_momentum;
	VECTOR<T> eulerian_mass;
	VECTOR<int> particle_cell_index;
	VECTOR<T> particle_weights;
	VECTOR_2D<T> gravity;
	std::string output_directory;
	int frames_per_second,frame,last_frame;
	VECTOR<bool> boundary_index;
	VECTOR<VECTOR_2D<T> > boundary_outward_direction;
	T two_thirds,one_sixth,four_thirds,one_over_dx;
	T epsilon,FLIP,PIC;
	T mu,lambda,youngs_modulus,poisson_ratio;
	
public:
	MPM_DRIVER(GRID_2D<T>& node_grid_input, VECTOR<T>& particle_position_input,VECTOR<T>& particle_velocity_input,
				  VECTOR<T>& particle_mass_input,VECTOR<T>& particle_area_input,const T dt_input,const T dx_particle_input):
	m(node_grid_input.M()),n(node_grid_input.N()),dx(node_grid_input.Dx()),
	dt(dt_input),time((T)0),dt_target(dt_input),dx_particle(dx_particle_input),CFL((T)0.01),
	node_grid(node_grid_input),
	particle_position(particle_position_input),
	particle_velocity(particle_velocity_input),
	particle_mass(particle_mass_input),
	particle_area(particle_area_input),
	F(particle_area_input.Size()),				//Define size of array containing matrices (each being a deformation gradient)
	FE(particle_area_input.Size()),
	FP(particle_area_input.Size()),
	youngs_modulus((T)2),poisson_ratio((T).3),
	eulerian_velocity(2*(node_grid_input.M()+1)*(node_grid_input.N()+1)),
	eulerian_velocity_old(2*(node_grid_input.M()+1)*(node_grid_input.N()+1)),
	eulerian_momentum(2*(node_grid_input.M()+1)*(node_grid_input.N()+1)),
	eulerian_mass((node_grid_input.M()+1)*(node_grid_input.N()+1)),
	particle_cell_index(particle_position.Size()),
	particle_weights(particle_position.Size()),
	gravity((T)0,-(T)1),
	output_directory("output/"),
	frames_per_second(30),frame(0),last_frame(330),
	boundary_index((node_grid_input.M()+1)*(node_grid_input.N()+1)),
	boundary_outward_direction((node_grid_input.M()+1)*(node_grid_input.N()+1)),
	two_thirds((T)2/(T)3),one_sixth((T)1/(T)6),four_thirds((T)4/(T)3),one_over_dx((T)1/node_grid_input.Dx()),
	epsilon((T)1e-8),FLIP((T).95),PIC((T)1-FLIP),
	thetaC( (T)2.4e-2 ),
	thetaS( (T)7.5e-3 ),
	hardening( (T)8 )
	{
		
		lambda=youngs_modulus*poisson_ratio/(((T)1+poisson_ratio)*((T)1-(T)2*poisson_ratio));
		mu=youngs_modulus/((T)2*((T)1+poisson_ratio));
		
		dt_per_frame=(T)1/(T)frames_per_second;
		//Set the dirichlet boundary layer
		boundary_index.Set_To_Zero();
		for(int i=0;i<node_grid.Number_Nodes_Per_X_Dimension();i++){
			for(int j=0;j<node_grid.Number_Nodes_Per_Y_Dimension();j++){	
				if(i<2 || i>node_grid.Number_Nodes_Per_X_Dimension()-3 || j<2 || j>node_grid.Number_Nodes_Per_Y_Dimension()-3)
					boundary_index(node_grid.Node_Index(i,j))=true;
				if(i==0 || i==1) boundary_outward_direction(node_grid.Node_Index(i,j))=VECTOR_2D<T>((T)1,(T)0);
				if(i==node_grid.Number_Nodes_Per_X_Dimension()-1 ||
				   i==node_grid.Number_Nodes_Per_X_Dimension()-2) 
					boundary_outward_direction(node_grid.Node_Index(i,j))=VECTOR_2D<T>(-(T)1,(T)0);
				if(j==0 || j==1) boundary_outward_direction(node_grid.Node_Index(i,j))=VECTOR_2D<T>((T)0,(T)1);
				if(j==node_grid.Number_Nodes_Per_Y_Dimension()-1 ||
				   j==node_grid.Number_Nodes_Per_Y_Dimension()-2)
					boundary_outward_direction(node_grid.Node_Index(i,j))=VECTOR_2D<T>((T)0,-(T)1);}}

		/////////////////////////// Initialize additional boundaries //////////////////////////////

		//Create a rombus in the center.
		int x = node_grid.Number_Nodes_Per_X_Dimension() / 2 - 15;	//Set the center of the rombus.
		int y = node_grid.Number_Nodes_Per_Y_Dimension() / 2 - 17;
		int width = 17;
		for( int depth = 0; depth < 5; depth++ )
		{
			std::vector< VECTOR_2D<int> > indices;					//Store rombus nodes indices here.
			std::vector< VECTOR_2D<T> > normals;					//Store rombus nodes normals here.
			Boundaries::createRombus(width, x, y, indices, normals);
			for( int r = 0; r < indices.size(); r++)				//This loop creates the rombus boundary.
			{
				VECTOR_2D<int> index = indices[r];					//Extract index.
				VECTOR_2D<T> normal = normals[r];					//Extract normal.
				boundary_index(node_grid.Node_Index(index(0), index(1))) = true;
				boundary_outward_direction(node_grid.Node_Index(index(0), index(1))) = normal;
			}

			width-=2;			//The inner rombus has span 2 less than the outer one.
		}

		///////////////////////////////////////////////////////////////////////////////////////////
		
		//Initialize deformation gradient and its elastic and plastic versions.
		for(int p=0;p<F.Size();p++)
		{
			F(p) = MATRIX_2X2<T>::Identity();
			FE(p) = MATRIX_2X2<T>::Identity();
			FP(p) = MATRIX_2X2<T>::Identity();
		}

		
		//compute initial particle weights
		Update_Weights();
		Transfer_Mass_From_Particles_To_Eulerian_Grid();
		Transfer_Momentum_From_Particles_To_Eulerian_Grid();
		Set_Eulerian_Velocity_From_Momentum(true);
		Write_Frame();
	}
	
	bool Is_Dirichlet(const int i, const int j){
		//helper for visualization of where dirichlet dofs are
		return boundary_index(node_grid.Node_Index(i,j));
	}
	
	void Set_Boundary_Index(INDEX_2D& index,bool is_boundary){
		//helper to set up dirichlet/collision dofs
		boundary_index(index.Index())=is_boundary;
	}

	/************ Function to get the normal of a dirichlet point **************/
	VECTOR_2D<T> Get_Dirichlet_Normal(const int i, const int j)
	{
		if( Is_Dirichlet( i, j ) )
			return VECTOR_2D<T>(boundary_outward_direction(node_grid.Node_Index(i,j)));
		else
			return VECTOR_2D<T>(0,0);
	}
	
	T Particle_Mass(const int particle_index){return particle_mass(particle_index);}
	VECTOR_2D<T> Particle_Position(const int particle_index){return VECTOR_2D<T>(particle_position(2*particle_index),particle_position(2*particle_index+1));}
	VECTOR_2D<T> Particle_Velocity(const int particle_index){return VECTOR_2D<T>(particle_velocity(2*particle_index),particle_velocity(2*particle_index+1));}
	VECTOR_2D<T> Particle_Position(const int particle_index,VECTOR<T>& x){return VECTOR_2D<T>(x(2*particle_index),x(2*particle_index+1));}
	VECTOR_2D<T> Particle_Velocity(const int particle_index,VECTOR<T>& v){return VECTOR_2D<T>(v(2*particle_index),v(2*particle_index+1));}
	VECTOR_2D<T> Eulerian_Velocity(const INDEX_2D node_index,VECTOR<T>& v){return VECTOR_2D<T>(v(2*node_index.Index()),v(2*node_index.Index()+1));}
	VECTOR_2D<T> Node(const INDEX_2D& index){return VECTOR_2D<T>(node_grid.X(index),node_grid.Y(index));}
	GRID_2D<T>& Node_Grid(){return node_grid;}
	int M()const{return m;}
	int N()const{return n;}
	
	T Total_Lagrangian_Mass(){
		T total_mass=(T)0;
		for(int p=0;p<particle_mass.Size();p++) total_mass+=particle_mass(p);
		return total_mass;
	}
	
	T Total_Eulerian_Mass(){
		T total_mass=(T)0;
		for(int ig=0;ig<eulerian_mass.Size();ig++) total_mass+=eulerian_mass(ig);
		return total_mass;
	}
	
	T Interpolating_Function(const T w){
		T abs_w=fabs(w);
		
		if(abs_w<=1){
			T abs_w_cubed=abs_w*abs_w*abs_w;
			T abs_w_squared=w*w;
			return (T).5*abs_w_cubed-abs_w_squared+two_thirds;}
		else if(abs_w>1 && abs_w<=2){
			T abs_w_cubed=abs_w*abs_w*abs_w;
			T abs_w_squared=w*w;
			return -one_sixth*abs_w_cubed+abs_w_squared-(T)2*abs_w+four_thirds;}
		return (T)0;
	}
	
	T Interpolating_Function_Derivative(const T w){
		T abs_w=fabs(w);
		
		if(abs_w<=1){
			return (T)1.5*abs_w*w-(T)2*w;}
		else if(abs_w>1 && abs_w<=2){
			return -(T).5*w*abs_w+(T)2*w-(T)2*w/abs_w;}
		return (T)0;
	}
	
	/* Function to compute the gradient of the interpolation function */
	/* Returns a 2D vector */
	VECTOR_2D<T> Grad_Nig(const T wx,const T wy){
		return one_over_dx*VECTOR_2D<T>(Interpolating_Function(wy)*Interpolating_Function_Derivative(wx),
										Interpolating_Function(wx)*Interpolating_Function_Derivative(wy));
	}
	
	T Maximum_Velocity(){
		T max_v=(T)0;
		for(int i=0;i<eulerian_velocity.Size()/2;i++){
			VECTOR_2D<T> v(eulerian_velocity(2*i),eulerian_velocity(2*i+1));
			max_v=std::max(max_v,v.Magnitude());}
		return max_v;
	}
	
	T Dt(){
		T max_v=Maximum_Velocity();
		T CFL_dt=dt_target;
		if(max_v>(T)0){
			CFL_dt=(CFL/max_v)*dx;
			if(CFL_dt>dt_target) CFL_dt=dt_target;}
		return CFL_dt;
	}
	
	/******* Main function to perform the simulation, it is excuted each time step ***********/
	void Advance_One_Time_Step(){
		
		if(frame<last_frame)
		{
			
			dt=Dt();
			bool write_frame=false;
			
			if(time+dt>=(T)(frame+1)*dt_per_frame)
			{
				dt=((T)(frame+1)*dt_per_frame)-time;
				frame++;
				write_frame=true;
			}
			
			Update_Eulerian_Momentum();						//Step 3: Update momentum in grid nodes.
			Set_Eulerian_Velocity_From_Momentum();			//Step 4: Compute velocities at the grid nodes.
			Update_Deformation_Gradient();					//Step 5: Update deformation gradient for all p particles.
			Update_Particle_Positions();					//Step 6 and 7: Update particles velocity and position.
			Update_Weights();
			Transfer_Mass_From_Particles_To_Eulerian_Grid();		//Step 1: Compute mass at grid nodes.
			Transfer_Momentum_From_Particles_To_Eulerian_Grid();	//Step 2: Compute momentum at the grid nodes.
			Set_Eulerian_Velocity_From_Momentum(true);		//Step 2.5: Before updating momentum next time step, store current velocities
															//but taking into account that they moved and their mass at the grid nodes
															//has changed.
			if(write_frame) Write_Frame();
			
			time+=dt;
		}
	}
	
	void Transfer_Mass_From_Particles_To_Eulerian_Grid(const bool debug=false){
	
		eulerian_mass.Set_To_Zero();
		for(int p=0;p<particle_position.Size()/2;p++){
			int icell=particle_cell_index(2*p);int jcell=particle_cell_index(2*p+1);
			int imin=icell-1;int jmin=jcell-1;int width=4;
			T w_x=particle_weights(2*p);T w_y=particle_weights(2*p+1);
			for(int is=0;is<width;is++){
				int ig=imin+is;
				for(int js=0;js<width;js++){
					int jg=jmin+js;
					eulerian_mass(node_grid.Node_Index(ig,jg))+=
					particle_mass(p)*Interpolating_Function((T)(1-is)+w_x)*Interpolating_Function((T)(1-js)+w_y);}}}
		
		if(debug){
			std::cout << "Total Lagrangian mass = " << Total_Lagrangian_Mass() << std::endl;
			std::cout << "Total Eulerian mass = " << Total_Eulerian_Mass() << std::endl;}
	}
	
	/********** Function to compute momentum at the grid nodes from the particles p **********/
	void Transfer_Momentum_From_Particles_To_Eulerian_Grid()
	{
		
		eulerian_momentum.Set_To_Zero();				//Initialize momentum.
		for(int p=0; p<particle_position.Size()/2; p++)	//Check each of the p particles.
		{
			int icell=particle_cell_index(2*p);			//Get the column coordinate (bottom left) of the cell where particle p is.
			int jcell=particle_cell_index(2*p+1);		//Get the row coordinate of the cell where particle p is.
			
			int imin=icell-1;							//Find the most bottom-left grid node that will be affected by particle p.
			int jmin=jcell-1;
			int width=4;								//Each p particle affects 16 grid nodes (because of the interpolation functions).
			
			T w_x=particle_weights(2*p);				//Get xai.
			T w_y=particle_weights(2*p+1);				//Get nu.
			
			for(int is=0; is<width; is++)				//Go through each grid node affected by particle p.
			{
				int ig = imin + is;
				for(int js=0;js<width;js++)				//Consider both coordinate components, e.g. (i,j).
				{
					int jg = jmin + js;
					T Nig = Interpolating_Function( (T)(1-is)+w_x ) * Interpolating_Function( (T)(1-js)+w_y );	//Compute interpolation val.
					
					//Update x component of momentum.
					eulerian_momentum(2*node_grid.Node_Index(ig,jg)) += particle_mass(p)*particle_velocity(2*p)*Nig;
					//Update y component of momentum.
					eulerian_momentum(2*node_grid.Node_Index(ig,jg)+1) += particle_mass(p)*particle_velocity(2*p+1)*Nig;
				}
			}
		}
	}
	
	/************** Function to compute the velocity at the grid nodes from the momentum ****************/
	void Set_Eulerian_Velocity_From_Momentum(const bool update_old_velocities=false)
	{
		eulerian_velocity.Set_To_Zero();						//Reset velocities at the grid nodes.
		
		for(int ig=0; ig<eulerian_velocity.Size()/2; ig++)		//Iterate over each grid node.
		{
			if(eulerian_mass(ig)>epsilon)						//Is there a significant amount of mass at this node?
			{
				eulerian_velocity(2*ig) = eulerian_momentum(2*ig)/eulerian_mass(ig);		//Compute x component for velocity.
				eulerian_velocity(2*ig+1) = eulerian_momentum(2*ig+1)/eulerian_mass(ig);	//Compute y component.
			}
		}
		
		//Traverse the grid to find out which nodes are at the boundaries in order to do something with grid nodes velocities
		//going outward the drawing plane.
		for(int i=0; i<node_grid.Number_Nodes_Per_X_Dimension(); i++)
		{
			for(int j=0; j<node_grid.Number_Nodes_Per_X_Dimension(); j++)
			{
				if(Is_Dirichlet(i,j))
				{
					int ig=node_grid.Node_Index(i,j);
					VECTOR_2D<T> normal=boundary_outward_direction(ig);
					T component=VECTOR_2D<T>::Dot_Product(normal,VECTOR_2D<T>(eulerian_velocity(2*ig),eulerian_velocity(2*ig+1)));
					if(component<(T)0)
					{
						eulerian_velocity(2*ig)-=component*normal.x();
						eulerian_velocity(2*ig+1)-=component*normal.y();
					}

					//Adding friction.
					eulerian_velocity(2*ig)*=0.25;
					eulerian_velocity(2*ig+1)*=0.25;
				}
			}
		}
		
		//If we need this velocity for FLIP, keep a copy of recently computed new velocities.
		if(update_old_velocities)
		{
			for(int ig=0; ig<eulerian_velocity.Size(); ig++) 
				eulerian_velocity_old(ig)=eulerian_velocity(ig);	//Copy element by element.
		}
	}
	
	/**************** Function to update momentum at the grid nodes *****************/
	void Update_Eulerian_Momentum(){
		
		//Add gravity
		for(int ig=0;ig<eulerian_velocity.Size()/2;ig++)
		{
			eulerian_momentum(2*ig)+=dt*eulerian_mass(ig)*gravity.x();
			eulerian_momentum(2*ig+1)+=dt*eulerian_mass(ig)*gravity.y();
		}
		
		//Add stress based forces, using deformation gradient updated last time step.
		for(int p = 0; p < particle_position.Size()/2; p++)
		{
			
			MATRIX_2X2<T> PiolaKirchoff = P(p);			//Compute Piola Kirchoff stress for p particle.
			MATRIX_2X2<T> PFt = PiolaKirchoff * FE(p).Transposed();

			//Update momentum of grid nodes affected by particle p.
			int icell = particle_cell_index( 2*p );		//Get the cell ith column index.
			int jcell = particle_cell_index( 2*p+1 );	//Get the cell jth row index.

			int imin = icell - 1;						//Most left bottom node to be affected by p.
			int jmin = jcell - 1;
			int width = 4;

			T xai = particle_weights( 2*p );			//Weight xai.
			T nu = particle_weights( 2*p+1 );			//Weight nu.

			for(int I = 0; I < width; I++)				//Iterate first over columns.
			{
				int ig = imin + I;						//The ig coordinate of the node to be updated.
				for(int J = 0; J < width; J++)			//Continue with rows.
				{
					int jg = jmin + J;					//The jg coordinate of the node to be updated.

					VECTOR_2D<T> gradNig = Grad_Nig((T)(1-I)+xai, (T)(1-J)+nu);	//Compute gradient of interpolating function.
					VECTOR_2D<T> forceAtP = PFt * gradNig;						//Compute force.

					eulerian_momentum( 2*node_grid.Node_Index(ig,jg) ) -= dt * forceAtP.x() * particle_area(p);	//Transfer x component.
					eulerian_momentum( 2*node_grid.Node_Index(ig,jg)+1 ) -= dt * forceAtP.y() * particle_area(p);	//Transfer y component.

				}
			}
		}

	}
	
	/************** Function to update particles velocity and position ********************/
	void Update_Particle_Positions()
	{
		
		//Interpolate velocity from grid nodes.
		for(int p=0; p<particle_position.Size()/2; p++)		//Do this for each particle p.
		{
			int icell=particle_cell_index(2*p);				//Find the bottom left grid coordinate of the cell where the particle is.
			int jcell=particle_cell_index(2*p+1);

			int imin=icell-1;				//Find the most bottom left node that will be affected by particle p.
			int jmin=jcell-1;
			int width=4;					//Also, particle p will affect 16 grid nodes.
			
			T w_x = particle_weights(2*p);	//xai weight.
			T w_y=particle_weights(2*p+1);	//nu weight.

			VECTOR_2D<T> v_new((T)0,(T)0);	//Initialize a new velocity for this particle p.
			VECTOR_2D<T> dv((T)0,(T)0);		//Change in velocity between previous momentum and current updated momentum.

			//Consider the 16 nodes that contribute to particle p.
			for(int is=0; is<width; is++)
			{
				int ig=imin+is;
				for(int js=0; js<width; js++)
				{
					int jg=jmin+js;

					T Nig=Interpolating_Function((T)(1-is)+w_x)*Interpolating_Function((T)(1-js)+w_y);	//Interpolation value for the location of
																										//particle p with respect to node (ig, jg).
					v_new.x() += Nig * eulerian_velocity(2*node_grid.Node_Index(ig,jg));	//PIC velocity components.
					v_new.y() += Nig * eulerian_velocity(2*node_grid.Node_Index(ig,jg)+1);

					//Compute the change in velocity between the new one and the old one.
					dv.x() += Nig*(eulerian_velocity(2*node_grid.Node_Index(ig,jg)) - eulerian_velocity_old(2*node_grid.Node_Index(ig,jg)));
					dv.y() += Nig*(eulerian_velocity(2*node_grid.Node_Index(ig,jg)+1) - eulerian_velocity_old(2*node_grid.Node_Index(ig,jg)+1));
				}
			}

			//Then, update velocity using an interpolation of PIC and FLIP.
			particle_velocity(2*p) = PIC*v_new.x() + FLIP*(particle_velocity(2*p)+dv.x());
			particle_velocity(2*p+1) = PIC*v_new.y() + FLIP*(particle_velocity(2*p+1)+dv.y());

			//Finally, update position.
			particle_position(2*p)+=dt*particle_velocity(2*p);
			particle_position(2*p+1)+=dt*particle_velocity(2*p+1);
		}
	}
	
	void Update_Weights(){
		for(int p=0;p<particle_position.Size()/2;p++){
			if(particle_position(2*p)<(node_grid.X_Min()+node_grid.Dx()+node_grid.Epsilon())){
				particle_position(2*p)=node_grid.X_Min()+node_grid.Dx()+node_grid.Epsilon();
				//if(particle_velocity(2*p)<(T)0) particle_velocity(2*p)=(T)0;
			}
			if(particle_position(2*p)>(node_grid.X_Max()-node_grid.Dx()-node_grid.Epsilon())){
				particle_position(2*p)=node_grid.X_Max()-node_grid.Dx()-node_grid.Epsilon();
				//if(particle_velocity(2*p)>(T)0) particle_velocity(2*p)=(T)0;
			}
			if(particle_position(2*p+1)<(node_grid.Y_Min()+node_grid.Dx()+node_grid.Epsilon())){
				particle_position(2*p+1)=node_grid.Y_Min()+node_grid.Dx()+node_grid.Epsilon();
				//if(particle_velocity(2*p+1)<(T)0) particle_velocity(2*p+1)=(T)0;
			}
			if(particle_position(2*p+1)>(node_grid.Y_Max()-node_grid.Dx()-node_grid.Epsilon())){
				particle_position(2*p+1)=node_grid.Y_Max()-node_grid.Dx()-node_grid.Epsilon();
				//if(particle_velocity(2*p+1)>(T)0) particle_velocity(2*p+1)=(T)0;
			}
			INDEX_2D node_index(0,0,node_grid.Number_Nodes_Per_X_Dimension(),node_grid.Number_Nodes_Per_Y_Dimension());
			T w_x,w_y;
			node_grid.Cell_Containing_Point(particle_position(2*p),particle_position(2*p+1),w_x,w_y,node_index);
			particle_weights(2*p)=w_x;particle_weights(2*p+1)=w_y;
			assert(node_index.I()>=0 && node_index.I()<=node_grid.Number_Nodes_Per_X_Dimension());
			assert(node_index.J()>=0 && node_index.J()<=node_grid.Number_Nodes_Per_Y_Dimension());
			particle_cell_index(2*p)=node_index.I();particle_cell_index(2*p+1)=node_index.J();}
	}
	
	/*************** Function to update the deformation gradient F ***************/
	void Update_Deformation_Gradient()
	{
		//Iterate over each particle p.
		for(int p = 0; p < particle_position.Size()/2; p++)
		{
			int icell = particle_cell_index(2*p);		//Find the bottom left grid coordinate of the cell where the particle is.
			int jcell = particle_cell_index(2*p+1);

			int imin = icell-1;				//Find the most bottom left node that will be affected by particle p.
			int jmin = jcell-1;
			int width = 4;					//Also, particle p will affect 16 grid nodes.

			T xai = particle_weights( 2*p );	//Get xai weight;
			T nu = particle_weights( 2*p+1 );	//Get nu.

			MATRIX_2X2<T> vTimesdNig( (T)0 );	//Cummulative matrix for the product of v and grad(Nig).

			for(int I = 0; I < width; I++)		//Move along columns.
			{
				int ig = imin + I;
				for(int J = 0; J < width; J++)	//Move along rows.
				{
					int jg = jmin + J;

					//Compute the gradient of the interpolating function at (ig,jg).
					VECTOR_2D<T> gradNig = Grad_Nig( (T)(1-I)+xai, (T)(1-J)+nu );
					
					//Get the velocity at (ig,jg).
					T xComponent = eulerian_velocity( 2*node_grid.Node_Index(ig, jg) );
					T yComponent = eulerian_velocity( 2*node_grid.Node_Index(ig, jg) + 1 );
					VECTOR_2D<T> vig( xComponent, yComponent );

					//Compute outer product.
					MATRIX_2X2<T> outerProduct = MATRIX_2X2<T>::Outer_Product( vig, gradNig );

					//Accumulate product.
					vTimesdNig += outerProduct;
				}
			}

			//Update deformation gradient at particle p.
			MATRIX_2X2<T> secondFactor = dt * vTimesdNig * F(p);
			F(p) += secondFactor;
		}

		//Now, update the elastic and plastic versions of the deformation gradient.
		for( int p = 0; p < particle_area.Size(); p++ )
		{
			//Compute the elastic version.
			MATRIX_2X2<T> UE, VE;
			VECTOR_2D<T> sigmaE;
			MATRIX_2X2<T> oldFPInverse = FP(p);		//Use old plastic deformation gradient.
			oldFPInverse.Invert();
			FE(p) = F(p) * oldFPInverse;			//Up to here, we have FE tilde.
			FE(p).SVD(UE, sigmaE, VE);				//Get the eigenvalues to clamp them.

			if( sigmaE(0) < (T)1-thetaC )			//Clamp first singular value.
				sigmaE(0) = (T)1-thetaC;
			else
			{
				if( sigmaE(0) > (T)1+thetaS )
					sigmaE(0) = (T)1+thetaS;
			}

			if( sigmaE(1) < (T)1-thetaC )			//Clamp second singular value.
				sigmaE(1) = (T)1-thetaC;
			else
			{
				if( sigmaE(1) > (T)1+thetaS )
					sigmaE(1) = (T)1+thetaS;
			}

			MATRIX_2X2<T> SigmaE( sigmaE );			//Construct the new matrix of eigenvalues.
			FE(p) = UE * SigmaE * VE.Transposed();	//This is the new elastic deformation gradient.

			//Compute the plastic version.
			MATRIX_2X2<T> FEInverse = FE(p);
			FEInverse.Invert();
			FP(p) = FEInverse * F(p);				//This is the new plastic deformation gradient.
		}

	}
	

	/*************** Function to compute the first Piola Kirchoff Stress *****************/
	MATRIX_2X2<T> P(const int p)
	{
		/*MATRIX_2X2<T> U_p,V_p;
		VECTOR_2D<T> sigma_p;
		F(p).SVD(U_p,sigma_p,V_p);
		MATRIX_2X2<T> JFinvT=F(p).Jacobian_Scaled_Inverse_Transpose();
		T J=sigma_p.x()*sigma_p.y();
		MATRIX_2X2<T> P_p=(T)2*mu*(F(p)-U_p*V_p.Transposed())+lambda*(J-1)*JFinvT;
		return P_p;*/

		//Do SVD decomposition on the elastic deformation gradient.
		MATRIX_2X2<T> U, V;
		VECTOR_2D<T> sigma;
		FE(p).SVD(U, sigma, V);

		//Define new mu and lambda.
		/*MATRIX_2X2<T> Up, Vp;
		VECTOR_2D<T> sigmap;
		FP(p).SVD(Up, sigmap, Vp);
		T JP = sigmap(0) * sigmap(1);*/					//Plastic Jacobian.
		T JP = FP(p).Determinant();
		T exponent = hardening * ((T)1 - JP);
		T newMu = mu * exp( exponent );
		T newLambda = lambda * exp( exponent );

		//Compute Piola-Kirchoff stress.
		MATRIX_2X2<T> JFEInvT = FE(p).Jacobian_Scaled_Inverse_Transpose();
		T JE = sigma(0) * sigma(1);
		MATRIX_2X2<T> Piola = (T)2 * newMu * (FE(p)-U*V.Transposed()) + newLambda * (JE - (T)1) * JFEInvT;

		//Return the Piola-Kirchoff stress for particle p.
		return Piola;
	}
	
	void Read_Domain_And_Frame_Info(int& frame_to_read,int& last_frame,int& m_input,int& n_input,T& dx_input,int &num_particles){
		std::ifstream indata;
		std::string simulation_data_filename(output_directory+std::string("simulation_info.dat"));
		indata.open(simulation_data_filename.c_str());
		indata >> last_frame;
		indata >> m_input;
		indata >> n_input;
		indata >> dx_input;
		indata >> num_particles;
		indata.close();
		if(frame_to_read>last_frame) frame_to_read=last_frame;
	}
	
	void Read_Frame(int frame_to_read,VECTOR<T>& particle_x,VECTOR<T>& particle_v,VECTOR<int>& particle_indices,VECTOR<T>& eulerian_v,VECTOR<T>& eulerian_m){
		
		assert(particle_x.Size()==particle_position.Size());
		assert(particle_v.Size()==particle_velocity.Size());
		assert(eulerian_v.Size()==2*(m+1)*(n+1));
		
		char str[12];
		sprintf(str, "%d", frame_to_read);	
		std::string frame_name(str);
		std::string eulerian_velocities_filename(std::string("eulerian_v_")+frame_name);
		READ_WRITE::Read_Binary(output_directory,eulerian_velocities_filename,eulerian_v);
		std::string eulerian_mass_filename(std::string("eulerian_mass_")+frame_name);
		READ_WRITE::Read_Binary(output_directory,eulerian_mass_filename,eulerian_m);
		std::string particle_positions_filename(std::string("particle_x_")+frame_name);
		READ_WRITE::Read_Binary(output_directory,particle_positions_filename,particle_x);
		std::string particle_velocities_filename(std::string("particle_v_")+frame_name);
		READ_WRITE::Read_Binary(output_directory,particle_velocities_filename,particle_v);
		std::string particle_indices_filename(std::string("particle_cell_index_")+frame_name);
		READ_WRITE::Read_Binary(output_directory,particle_indices_filename,particle_indices);
	}
	
	void Write_Frame(){
		
		std::ofstream outdata;
		std::string simulation_data_filename(output_directory+std::string("simulation_info.dat"));
		outdata.open(simulation_data_filename.c_str());
		outdata << frame << std::endl;
		outdata << m << std::endl;
		outdata << n << std::endl;
		outdata << dx << std::endl;
		outdata << particle_position.Size()/2;
		outdata.close();
		
		char str[12];
		sprintf(str, "%d", frame);	
		std::string frame_name(str);
		std::string eulerian_velocities_filename(std::string("eulerian_v_")+frame_name);
		READ_WRITE::Write_Binary(output_directory,eulerian_velocities_filename,eulerian_velocity);
		std::string eulerian_mass_filename(std::string("eulerian_mass_")+frame_name);
		READ_WRITE::Write_Binary(output_directory,eulerian_mass_filename,eulerian_mass);
		std::string particle_positions_filename(std::string("particle_x_")+frame_name);
		READ_WRITE::Write_Binary(output_directory,particle_positions_filename,particle_position);
		std::string particle_velocities_filename(std::string("particle_v_")+frame_name);
		READ_WRITE::Write_Binary(output_directory,particle_velocities_filename,particle_velocity);
		std::string particle_indices_filename(std::string("particle_cell_index_")+frame_name);
		READ_WRITE::Write_Binary(output_directory,particle_indices_filename,particle_cell_index);
	}
	
};
#endif